#ifndef __VBIOS_P_C_INCLUDED
#define __VBIOS_P_C_INCLUDED 1

#include "vbios.h"

/****************************************************************************
 *
 * System (global) variables
 *
 ****************************************************************************/

/* the pointer is a constant value.  But the referenced structure is NOT
   constant!  Position of the term 'const' is important! 
 */

 physical_display_descriptor * const g_pdd = (physical_display_descriptor *)
	VBIOS_PDD_ADDRESS; /* the system reserved pointer to pdd */

 viewport_display_descriptor * const g_vdd = (viewport_display_descriptor *)
	VBIOS_VDD_ADDRESS; /* the system reserved pointer to vdd */

/****************************************************************************
 *
 * physical display routines
 *
 * Physical routines directly manipulate video controller RAM, or registers.
 *
 ****************************************************************************/

signed int 
vbios_validate_physical_coordinates( const unsigned int tX,
	const unsigned int tY )
{
	/* level 1 check : obvious out of bound parameters */
	if ( (tX < 0) || (tY < 0) || (tX >= g_pdd->c_tWidth) ||
		(tY >= g_pdd->c_tHeight) )
		return VBIOS_RC_FAIL;
	else
		return VBIOS_RC_OK;
}


signed int 
vbios_validate_physical_graphics_coord( const unsigned int gX, 
	const unsigned int gY )
{
	/* level 1 check : obvious out of bound parameters */
	if ( (gX < 0) || (gY < 0) || (gX >= g_pdd->w_gWidth) ||
		(gY >= g_pdd->w_gHeight) )
		return VBIOS_RC_FAIL;
	else
		return VBIOS_RC_OK;
}


signed int 
vbios_p_get_char( const unsigned int tX, const unsigned int tY )
{
	unsigned char *tptr; /* pointer to text buffer */

	/* validate tX, tY */
	if ( !(VBIOS_SUCCEED(vbios_validate_physical_coordinates( tX, tY ))) )
		return VBIOS_RC_FAIL;

	/* get pointer to top-left (0,0) of text-screen framebuffer */
	tptr = (unsigned char *) (g_pdd->pTBase);  

	/* set mptr to text_grid(tX,tY) */
	tptr = tptr + ((g_pdd->w_tPitch) * tY) + tX;

	/* return character code from text framebuffer */
	return (*tptr);
}


signed int 
vbios_p_draw_gchar( const unsigned int gX, const unsigned int gY, 
				const unsigned char ccode )
{
	unsigned char *fptr; /* font bitmap pointer */
	unsigned char *mptr; /* framebuffer pointer */
	unsigned char temp_l, temp_r, mask_l, mask_r; /* temp scratch register */
	unsigned char ftemp, ftemp_l, ftemp_r;
	unsigned int y, ymax;
	unsigned int tX, deltaX;

	/* validate tX, tY */
	if ( !(VBIOS_SUCCEED( vbios_validate_physical_graphics_coord( gX, gY ))) )
		return VBIOS_RC_FAIL;

	/* plotting process : 
	 *
	 * The plot 
	 *               |<------------->|
	 *               gY
	 *    |       n       |      n+1      |
	 *    |7 6 5 4 3 2 1 0|7 6 5 4 3 2 1 0|  <- BYTE memory-address
	 *    |<-------->|    |               |
	 *      deltaX=5      |
	 *               |<-->|
	 *              8-deltaX
	 */


	/* calculate the leftmost character period in plot area */
	tX = gX >> 3;
	deltaX = gX & 0x07; /* left pixeloffset */

	/* check whether gX is byte-aligned, if so call 'fast-draw' */
	if ( deltaX == 0 )
		return vbios_p_draw_gchar_fast( tX, gY, ccode );

	/* set mptr to top-left pixel of target-draw area */
	mptr = (unsigned char *) (g_pdd->pGBase);
	mptr = mptr + (g_pdd->w_gPitch * gY) + tX;

	/* set fptr to start of font_bitmap[ ccode ] */
	fptr = (unsigned char *) (g_pdd->pFBase);
	fptr = fptr + (g_pdd->c_fHeight * ccode); 

	mask_l = (0xFFU) << (8-deltaX);/* shift in zeros from left*/
	mask_r = (0xFFU) >> deltaX;   /* shift in zeros from right */

	/* ymax = min( g_pdd->c_fHeight, g_pdd->w_gHeight-gY ) */
	/* ...to clip at bottom of screen */
	if ( (gY + g_pdd->c_fHeight) > g_pdd->w_gHeight )
	{
		/* gY is near bottom of screen; cut off bottom of character */
		if ( g_pdd->w_gHeight < gY )
			ymax = 0; /* don't plot character at all */
		else 
			ymax = g_pdd->w_gHeight - gY;
	}
	else
		ymax = g_pdd->c_fHeight; /* plot full height of character */

	for ( y = 0; y < ymax; ++y)
	{
		/* 1) get the font bitmap slice */
		ftemp = *(fptr + y );
		ftemp_l = ftemp >> deltaX;
		ftemp_r = ftemp << (8-deltaX);

		/* 2) Do a read & modify */
		temp_l = *(mptr + g_pdd->w_gPitch * y );
		temp_l = temp_l & mask_l;

		/* 3) Now write result back to framebuffer */
		*(mptr + g_pdd->w_gPitch * y ) = temp_l | ftemp_l;

		/* Now repeat (1)-(3) for right-part of char, if needed */
		if ( deltaX != 0 && ((gX+8) < g_pdd->w_gWidth) )
		{
			temp_r = *(mptr + g_pdd->w_gPitch * y + 1); 
			temp_r = temp_r & mask_r;
			*(mptr + g_pdd->w_gPitch * y + 1) = temp_r | ftemp_r;
		}
	}

	return VBIOS_RC_OK;
}


signed int 
vbios_p_draw_gchar_fast( const unsigned int tX, const unsigned int gY, 
				const unsigned char ccode )
{
	unsigned char *fptr; /* font bitmap pointer */
	unsigned char *mptr; /* framebuffer pointer */
	unsigned char  ftemp;
	unsigned int y;
	unsigned int gX;

	/* convert (tX, tY) into pixel coordinates (gX, gY) */
	gX = tX << 3;

	if ( !(VBIOS_SUCCEED( vbios_validate_physical_graphics_coord( gX, gY ))) )
		return VBIOS_RC_FAIL;

	/* set mptr to top-left pixel of target-draw area */
	mptr = (unsigned char *) (g_pdd->pGBase);
	mptr = mptr + (g_pdd->w_gPitch * gY) + tX;

	/* set fptr to start of font_bitmap[ ccode ] */
	fptr = (unsigned char *) (g_pdd->pFBase);
	fptr = fptr + (g_pdd->c_fHeight * ccode); 

	for ( y = 0; y < g_pdd->c_fHeight; ++y)
	{
		/* get the font bitmap slice */
		ftemp = *(fptr + y );

		/* Now write result back to framebuffer */
		*(mptr + g_pdd->w_gPitch * y ) = ftemp;
	}

	return VBIOS_RC_OK;
}


signed int 
vbios_p_put_char( const unsigned int tX, const unsigned int tY,
	const unsigned char ccode )
{
	unsigned char *tptr; /* textbuffer pointer */

	/* validate tX, tY */
	if ( !(VBIOS_SUCCEED( vbios_validate_physical_coordinates( tX, tY ))) )
		return VBIOS_RC_FAIL;

	/* get pointer to top-left (0,0) of text-screen framebuffer */
	tptr = (unsigned char *) (g_pdd->pTBase);  

	/* set mptr to text_grid(tX,tY) */
	tptr = tptr + ((g_pdd->w_tPitch) * tY) + tX;

	/* write the ccode to text framebuffer */
	(*tptr) = ccode;


	return VBIOS_RC_OK;
}


/****************************************************************************
 *
 * text viewport routines
 *
 ****************************************************************************/


signed int 
vbios_validate_viewport_coordinates( const unsigned int tX, 
									const unsigned int tY )
{
	unsigned int pX, pY; /* physical text coordinates */

	/* level 1 check : obvious out of bound parameters */
	if ( (tX < 0) || (tY < 0) || (tX >= g_vdd->c_tWidth) ||
		(tY >= g_vdd->c_tHeight) )
		return VBIOS_RC_ILLEGAL_PARAMS;

	/* level 2 check : is input coordinate in visible space  */

	/* convert viewport (logical) coordinate -> physical (screen) coords */
	pX = tX + g_vdd->c_tOffsetX;
	pY = tY + g_vdd->c_tOffsetY;

	if ( VBIOS_SUCCEED( vbios_validate_physical_coordinates( pX, pY)) )
		return VBIOS_RC_OK;
	else
		return VBIOS_RC_NOT_VISIBLE; /* outside physical screen */
}


signed int 
vbios_v_get_char( const unsigned int tX, const unsigned int tY )
{
	unsigned int pX, pY; /* physical text coordinates */

	/* validate tX, tY */
	if ( !(VBIOS_SUCCEED(vbios_validate_viewport_coordinates( tX, tY ))) )
		return VBIOS_RC_FAIL;

	/* convert viewport (logical) coordinate -> physical (screen) coords */
	pX = tX + g_vdd->c_tOffsetX;
	pY = tY + g_vdd->c_tOffsetY;


	return vbios_p_get_char( pX, pY );
}


signed int 
vbios_v_put_char( const unsigned int tX, const unsigned int tY, 
				 const unsigned char ccode )
{
	unsigned int pX, pY; /* physical text coordinates */
	unsigned int gX, gY; /* physical graphics coordinates */

	/* validate tX, tY */
	if ( !(VBIOS_SUCCEED(vbios_validate_viewport_coordinates( tX, tY ))) )
		return VBIOS_RC_FAIL;

	/* convert viewport (logical) coordinate -> physical (screen) coords */
	pX = tX + g_vdd->c_tOffsetX;
	pY = tY + g_vdd->c_tOffsetY;

	gX = (pX << 3) + g_pdd->w_gOffsetX; 
	gY = ( pY * g_pdd->c_fHeight ) + g_pdd->w_gOffsetY;

	/* Optional, if in graphics mode (pixel-addressable), write bitmap */
	if ( VBIOS_IS_MODE_GRAPHICS( g_pdd->w_mode ) )
	{
		vbios_p_draw_gchar( gX, gY, ccode );
	}

	return vbios_p_put_char( pX, pY, ccode);
}

signed int 
vbios_v_move_cursor( const unsigned int tX, const unsigned int tY )
{
	/* validate tX, tY */
	if ( !(VBIOS_SUCCEED(vbios_validate_viewport_coordinates( tX, tY ))) )
		return VBIOS_RC_FAIL;

	/* set cursor position to requested location (tX,tY) */
	g_vdd->c_tCursorXpos = tX;
	g_vdd->c_tCursorYpos = tY;

	return VBIOS_RC_OK;
}

signed int 
vbios_v_scroll_screen( const unsigned int rowCount, const unsigned int direction,
					const unsigned char fillcode )
{
	unsigned int x, y; /* loop variables */
	unsigned int value; /* temp value */
	unsigned char ccode; /* temp value */
	signed int rc; /* return code */

	/* validate rowCount */
	if ( rowCount < 0 || rowCount > g_vdd->c_tHeight )
		return VBIOS_RC_ILLEGAL_PARAMS;
	
	if ( direction == VBIOS_SCROLL_UP )
	{
		/* move visible screen upward */
		for ( y = rowCount; y < g_vdd->c_tHeight; ++y )
			for ( x = 0; x < g_vdd->c_tWidth; ++x )
			{
				value = vbios_v_get_char(x, y );
				ccode = value;
				if ( VBIOS_GET_CHAR_SUCCEED( value ) ) 
					vbios_v_put_char(x, (y-rowCount), ccode);
			} /* end for ( x */
		/* end for ( y */

		/* fill the added lines (at bottom) with fillcode */
		for ( y=1; y <= rowCount; ++y )
			for ( x = 0; x < g_vdd->c_tWidth; ++x )
			{
				rc = vbios_v_put_char(x, 
					g_vdd->c_tHeight - y, fillcode );
			}

	}
	else if ( direction == VBIOS_SCROLL_DOWN )
	{
		/* move visible screen downward */
		y = (g_vdd->c_tHeight) - 1;
		for ( ; y >= rowCount; --y )
			for ( x = 0; x < g_vdd->c_tWidth; ++x )
			{
				value = vbios_v_get_char(x, (y-rowCount));
				ccode = value;
				if ( VBIOS_GET_CHAR_SUCCEED( value ) ) 
					vbios_v_put_char(x, y, ccode);
			} /* end for ( x */
		/* end for ( y */

		/* fill the inserted lines (at top) with fillcode */
		for ( y = 0; y < rowCount ; ++y )
			for ( x = 0; x < g_vdd->c_tWidth; ++x )
			{
				rc = vbios_v_put_char(x, y, fillcode );
			}
	} 
	else return VBIOS_RC_ILLEGAL_PARAMS; /* unrecognized scroll code */

	return VBIOS_RC_OK;
}


signed int 
vbios_v_scroll_cursor( const int deltaX, const int deltaY )
{
	signed int rc;
	signed int currentX = (unsigned int) g_vdd->c_tCursorXpos;
	signed int currentY = (unsigned int) g_vdd->c_tCursorYpos;
	signed int ndeltaX, ndeltaY;
	signed int step;
	unsigned int newX, newY;

	ndeltaX = deltaX;
	ndeltaY = deltaY;

	/* first, update X, increment +/-1 per loop-iteration */
	for ( ndeltaX = deltaX; ndeltaX != 0; ndeltaX -= step )
	{
		if ( ndeltaX < 0 )
			step = -1;/* move left */
		else if ( ndeltaX > 0 )
			step = 1; /* move right */
		else
			step = 0; /* don't move */

		/* update cursor position */
		currentX = currentX + step; 

		/* check for left-space before start-of-line */
		if ( (currentX < 0) && (step == -1) ) 
		{	
			/* cursor is at 'start-of-line', wrap up one line */
			currentX = g_vdd->c_tWidth - 1; /* set to end-of-line */
			ndeltaY = ndeltaY - 1; /* move up 1 line */
		}

		/* check for right-space past end-of-line */
		if ( (currentX >= g_vdd->c_tWidth) && step==1 ) 
		{	
			/* cursor is at 'end-of-line', simulate a carriage-return */
			currentX = 0; /* set to start-of-line */
			ndeltaY = ndeltaY + 1; /* move down 1 line */
		}

	} /* for ( ndeltaX = deltaX; ... */

	/* second, update Y, increment +/-1 per loop-iteration  */
	for( ; ndeltaY != 0; ndeltaY -= step )
	{
		if ( ndeltaY < 0 )
			step = -1;/* move up */
		else if ( ndeltaY > 0 )
			step = 1; /* move down */
		else
			step = 0; /* don't move */

		/* update cursor position */
		currentY = currentY + step; 

		/* check for scroll-up beyond top-of-screen */
		if ( (currentY < 0) && (step == -1) ) 
		{	
			/* cursor is at 'top-of-screen', must insert lines */
			currentY = 0;
			vbios_v_scroll_screen( 1, VBIOS_SCROLL_DOWN, ASCII_SPACE );
		}
		
		/* check for scroll-down beyond bottom-of-screen */
		if ( (currentY >= g_vdd->c_tHeight) && step==1 ) 
		{	
			/* cursor is at 'bottom-of-screen', must insert lines */
			currentY = g_vdd->c_tHeight - 1;
			vbios_v_scroll_screen( 1, VBIOS_SCROLL_UP, ASCII_SPACE );
		}
	} /* for ( ndeltaY ... */

	/* update viewport data structure */
	newX = (unsigned int) currentX;
	newY = (unsigned int) currentY;
	rc = vbios_v_move_cursor( newX, newY );

	return VBIOS_RC_OK;
}


signed int 
vbios_v_write_char( const unsigned char ccode )
{
	signed int rc;
	signed int deltaX, deltaY; /* grid displacements */
	unsigned int bCharIsVisible; /* flag, character is visible? */

	/* calculate displacement from current cursor position */
	deltaX = 0;
	deltaY = 0;
	bCharIsVisible = TRUE;

	if ( ccode == ASCII_BACKSPACE )
	{
		deltaX = -1; /* retreat one space to the left */
		bCharIsVisible = FALSE;
	}
	else if ( ccode == ASCII_CR ) 
	{
		/* CR = return cursor to LEFT of line (don't advance down!) */
		deltaX = (g_vdd->c_tCursorXpos); /* backstep to start-of-line */
		deltaX = -deltaX; /* avoid LCC complaint */
		bCharIsVisible = FALSE;
	}
	else if ( ccode == ASCII_LF )
	{
		deltaY = 1; /* advance one line down */
		bCharIsVisible = FALSE;
	}
	else 
	{
		/* ordinary visible ASCII character */
		deltaX = 1; /* advance one space to the right */
		bCharIsVisible = TRUE;
	}

	if ( bCharIsVisible == TRUE )
	{
		rc = vbios_v_put_char( g_vdd->c_tCursorXpos, g_vdd->c_tCursorYpos, 
			ccode );
		if ( !VBIOS_SUCCEED( rc ) )
			return rc; /* operation failed, exit early */
	}

	/* update the cursor position */
	rc = vbios_v_scroll_cursor( deltaX, deltaY );

	return rc;	
}

signed int 
vbios_v_write_string( const char cstring[] )
{
	unsigned int i;
	signed int rc = VBIOS_RC_OK;

	i = 0;
	while ( i < 4096U && VBIOS_SUCCEED( rc ) && cstring[i] )
	{
		 rc = vbios_v_write_char( cstring[i] );
		 ++i;
	}


	return rc;
}

signed int 
vbios_p_init( void )
{
/*	g_pdd = ( physical_display_descriptor *) VBIOS_PDD_ADDRESS;*/

	g_pdd->w_gHeight = 448; /* graphics screen height (#pixels) */
	g_pdd->w_gWidth = 560; /* graphics screen width (#pixels) */
	g_pdd->c_fHeight = 16; /* font character-cell height (#pixels) */
	g_pdd->c_fWidth = 8; /* font character-cell width (#pixels) */
	g_pdd->c_tHeight = 28; /* text screen height (#cells) */
	g_pdd->c_tWidth = 70; /* text screen width (#cells) */

	/* pointer to (0,0) of graphics screen framebuffer */
	/* 640x480 = 37.5KB, 4A00 - DFFF */
	vbios_set_framebuffer( VBIOS_MK_PTR( 0x4A00U ) );

	/* pointer to (0,0) of text screen framebuffer */
	/* 80x51 text screen = 4KB, 3A00-49FF */
	vbios_set_textbuffer( VBIOS_MK_PTR( 0x3A00U ) );

	/* pointer to (0,0) of font bitmap */
	/* 256 chars * 8x31 = 7.75KB font */
	vbios_set_font_address( VBIOS_MK_PTR( 0xE000U ) );

	/* set display physical origin to (0,0) pixels */
	vbios_set_gOffsetX( 0U );
	vbios_set_gOffsetY( 0U ); 

	g_pdd->w_gPitch = 72; /* pitch of graphics screen framebuffer (#bytes) */
	g_pdd->w_tPitch = 72; /* row pitch of text screen framebuffer (#bytes) */

	g_pdd->w_mode = VBIOS_MODE_GRAPHICS1; /* not used */

	return VBIOS_RC_OK;
}

signed int 
vbios_v_init( void )
{
	/* initialize the physical display */
	vbios_p_init( );

/*	g_vdd = ( viewport_display_descriptor *) VBIOS_VDD_ADDRESS;*/

	g_vdd->c_tHeight = 28;
	g_vdd->c_tWidth = 70; 

	/* X offset from base-coordinate (0,0), #cells */
	/* Y offset from base-coordinate (0,0), #cells */
	g_vdd->c_tOffsetX = 0; 
	g_vdd->c_tOffsetY = 0; 

	/* cursor position, X-coordinate (in viewport co-ordinate space) */
	/* cursor position, Y-coordinate (in viewport co-ordinate space) */
	g_vdd->c_tCursorXpos = 0; 
	g_vdd->c_tCursorYpos = 0; 
	/* ASCII character-code of the cursor blinker */
	g_vdd->c_tCursorCode = 254; 

	/* flag, 1=cursor is visible, 0=cursor invisible */
	g_vdd->w_tCursorVisible = TRUE; 

	/* pointer to physical-descriptor data-structure */
	g_vdd->pPDescriptor = ( physical_display_descriptor *) g_pdd; 

	/* clear the viewport area */
	vbios_v_scroll_screen( g_vdd->c_tHeight, 
						VBIOS_SCROLL_UP, ASCII_SPACE );

	return VBIOS_RC_OK;
}


signed int 
vbios_set_font_address( const unsigned char font[] )
{
	g_pdd->pFBase = VBIOS_MK_PTR( font );

	return VBIOS_RC_OK;
}


signed int 
vbios_set_framebuffer( const unsigned char framebuffer[] )
{
	g_pdd->pGBase = VBIOS_MK_PTR( framebuffer );

	return VBIOS_RC_OK;
}


signed int 
vbios_set_textbuffer( const unsigned char textbuffer[] )
{
	g_pdd->pTBase = VBIOS_MK_PTR( textbuffer );

	return VBIOS_RC_OK;
}

signed int 
vbios_set_gOffsetX( const unsigned int offset )
{
	if ( offset > 7 )
		return VBIOS_RC_ILLEGAL_PARAMS;
	else
		g_pdd->w_gOffsetX = offset;

	return VBIOS_RC_OK;
}

signed int 
vbios_set_gOffsetY( const unsigned int offset )
{
	if ( offset > 0 )
		return VBIOS_RC_ILLEGAL_PARAMS;
	else
		g_pdd->w_gOffsetY = offset;

	return VBIOS_RC_OK;

}

unsigned int 
vbios_get_gOffsetX( void )
{
	return g_pdd->w_gOffsetX;
}

unsigned int 
vbios_get_gOffsetY( void )
{
	return g_pdd->w_gOffsetY;
}

#endif

